use crate::application;
use gtk4 as gtk;
use gtk::{glib, gio, gdk, prelude::*, subclass::prelude::*};

mod imp {
    use super::*;
    use glib::clone;
    use std::cell::RefCell;
    use glib::subclass::Signal;
    use once_cell::sync::Lazy;
    use simplelogin::reqwest;
    
    #[derive(Debug, Default, glib::Properties, gtk::CompositeTemplate)]
    #[template(file = "data/settings_panel.blp")]
    #[properties(wrapper_type = super::SettingsPanel)]
    pub struct SettingsPanel {
        #[template_child]
        pub overlay: TemplateChild<gtk::Overlay>,
        #[template_child]
        pub content_wrapper: TemplateChild<gtk::Box>,
        #[template_child]
        pub content: TemplateChild<gtk::Grid>,
        #[template_child]
        pub profile_image: TemplateChild<gtk::Image>,
        #[template_child]
        pub name_label: TemplateChild<gtk::Entry>,
        #[template_child]
        pub edit_button: TemplateChild<gtk::Button>,
        #[template_child]
        pub cancel_edit_button: TemplateChild<gtk::Button>,
        #[template_child]
        pub stats_label: TemplateChild<gtk::Label>,
        #[template_child]
        pub email_label: TemplateChild<gtk::Label>,
        #[template_child]
        pub logout_button: TemplateChild<gtk::Button>,
        #[template_child]
        pub api_key_field: TemplateChild<gtk::PasswordEntry>,
        #[template_child]
        pub status_label: TemplateChild<gtk::Label>,
        #[template_child]
        pub api_key_clear_button: TemplateChild<gtk::Button>,
        #[template_child]
        pub loading_overlay: TemplateChild<gtk::Box>,
        #[template_child]
        pub error_overlay: TemplateChild<gtk::Box>,
        #[template_child]
        pub error_refresh_button: TemplateChild<gtk::Button>,
        #[template_child]
        pub error_close_button: TemplateChild<gtk::Button>,
        #[template_child]
        pub error_label: TemplateChild<gtk::Label>,
        #[property(get, construct_only)]
        pub application: RefCell<application::Application>,
    }

    #[gtk::template_callbacks]
    impl SettingsPanel {
        fn set_is_loading(&self, loading: bool) {
            if loading {
                self.set_error(None);
            } else if !self.is_error() {
                self.content_wrapper.get().set_sensitive(true);
            }
            self.loading_overlay.get().set_visible(loading);
        }

        fn is_error(&self) -> bool {
            self.error_overlay.get().is_visible()
        }

        fn set_error(&self, error: Option<&str>) {
            if let Some(e) = error {
                self.set_is_loading(false);
                self.error_label.get().set_text(e);
            }
            self.content_wrapper.get().set_sensitive(error.is_none());
            self.error_overlay.get().set_visible(error.is_some());
        }

        #[template_callback]
        fn clear_api_key_clicked(&self, _: &gtk::Button) {
            let main_loop = glib::MainContext::default();
            main_loop.spawn_local(clone!(@weak self as this => async move {
                let alert_diag = gtk::AlertDialog::builder()
                    .message("Are you sure you would like to clear the API key?")
                    .buttons(["Yes", "No"]).build();
                let widget_this = application::get_window_ancestor(
                    &this.obj().clone().upcast::<gtk::Widget>());
                let resp = alert_diag.choose_future(widget_this.as_ref()).await;
                if let Ok(resp_num) = resp {
                    if resp_num == 0 {
                        let app = this.application.borrow();
                        app.context().borrow_mut().set_api_key(None);
                        this.obj().emit_by_name::<()>("api-key-cleared", &[]);
                    }
                }
            }));
        }

        #[template_callback]
        fn logout_clicked(&self, _: &gtk::Button) {
            let main_loop = glib::MainContext::default();
            main_loop.spawn_local(clone!(@weak self as this => async move {
                let alert_diag = gtk::AlertDialog::builder()
                    .message("Are you sure you would like to log out?")
                    .buttons(["Yes", "No"]).build();
                let widget_this = application::get_window_ancestor(
                    &this.obj().clone().upcast::<gtk::Widget>());
                let resp = alert_diag.choose_future(widget_this.as_ref()).await;
                if let Ok(resp_num) = resp {
                    if resp_num == 0 {
                        let app = this.application.borrow();
                        let mut context = app.context().borrow_mut();
                        let res = context.logout().await;
                        drop(context);
                        match res {
                            Ok(_) => this.obj().emit_by_name::<()>("logged-out", &[]),
                            Err(err) => this.set_error(Some(err.to_string().as_str())),
                        }
                    }
                }
            }));
        }

        #[template_callback]
        fn close_error_clicked(&self, _: &gtk::Button) {
            self.set_error(None);
        }

        async fn refresh_stats(&self) -> bool {
            let app = self.application.borrow();
            let mut ctx = app.context().borrow_mut();
            match ctx.stats().await {
                Ok(stats) => {
                    self.stats_label.get().set_text(&format!(
                        "Aliases: {}, Blocked: {}, Forwarded: {}, Replied: {}",
                        stats.nb_alias, stats.nb_block, stats.nb_forward, stats.nb_reply
                    ));
                    false
                },
                Err(e) => {
                    self.set_error(Some(e.to_string().as_str()));
                    true
                },
            }
        }

        fn did_profile_picture_change(
            new_url: &Option<String>, settings: &gio::Settings
        ) -> bool {
            let old_url: Option<String> = settings.value("profile-url")
                .get::<Option<String>>().unwrap();
            *new_url != old_url
        }

        fn has_valid_content_type(resp: &reqwest::Response) -> bool {
            match resp.headers().get("Content-Type") {
                Some(ct) => match ct.to_str() {
                    Ok("image/png") => true,
                    Ok("image/jpeg") => true,
                    _ => false,
                },
                None => false,
            }
        }

        async fn process_profile_picture_response_body(
            &self, url: &str, resp: reqwest::Response,
            settings: &gio::Settings
        ) {
            if let Ok(body) = resp.bytes().await {
                let tex_res = gdk::Texture::from_bytes(&glib::Bytes::from(body.as_ref()));
                if let Ok(tex) = tex_res {
                    self.profile_image.get().set_from_paintable(Some(&tex));
                    _ = settings.set("profile-url", Some(url));
                    _ = settings.set("profile-data", Some(body.as_ref()));
                }
            }
        }

        async fn download_and_set_profile_picture(
            &self, url: &str, app: &application::Application
        ) -> bool {
            let ctx = app.context().borrow();
            let client = ctx.client();
            match client.get(url).send().await {
                Ok(resp) if Self::has_valid_content_type(&resp) => {
                    self.process_profile_picture_response_body(
                        url, resp, &app.settings()).await;
                    false
                },
                Ok(_) => false,
                Err(e) => {
                    self.set_error(Some(&e.to_string()));
                    true
                },
            }
        }

        async fn refresh_user_profile_picture(
            &self, user_info: &simplelogin::UserInfo,
        ) -> bool {
            let app = self.application.borrow();
            let settings = &app.settings();
            if Self::did_profile_picture_change(&user_info.profile_picture_url,
                &settings) {
                    self.profile_image.get()
                        .set_from_icon_name(Some("user-info-symbolic"));
                    _ = settings.set("profile-url", Option::<&str>::None);
                    _ = settings.set("profile-data", Option::<&[u8]>::None);
                    match &user_info.profile_picture_url {
                        Some(new_url) => self.download_and_set_profile_picture(
                            &new_url, &app).await,
                        None => false,
                    }
                } else {
                    false
                }
        }

        async fn refresh_user_info(&self) -> bool {
            let app = self.application.borrow();
            let mut ctx = app.context().borrow_mut();
            match ctx.get_user_info().await {
                Ok(user_info) => {
                    self.name_label.get().set_text(
                        if user_info.name.is_empty() {
                            "<No name>"
                        } else {
                            &user_info.name
                        }
                    );
                    self.email_label.get().set_text(&user_info.email);
                    self.status_label.get().set_text(
                        if user_info.is_premium {
                            "Premium"
                        } else if user_info.in_trial {
                            "Premium (Trial)"
                        } else {
                            "Free"
                        });
                    drop(ctx);
                    self.refresh_user_profile_picture(&user_info).await;
                    false
                },
                Err(e) => {
                    self.set_error(Some(e.to_string().as_str()));
                    true
                },
            }
        }
    }

    #[glib::object_subclass]
    impl ObjectSubclass for SettingsPanel {
        const NAME: &'static str = "SettingsPanel";
        type Type = super::SettingsPanel;
        type ParentType = gtk::Box;

        fn class_init(class: &mut Self::Class) {
            class.bind_template();
            class.bind_template_callbacks();
        }

        fn instance_init(obj: &glib::subclass::InitializingObject<Self>) {
            obj.init_template();
        }
    }

    #[glib::derived_properties]
    impl ObjectImpl for SettingsPanel {
        fn signals() -> &'static [Signal] {
            static SIGNALS: Lazy<Vec<Signal>> = Lazy::new(||
                vec![Signal::builder("logged-out").build(),
                    Signal::builder("api-key-cleared").build()]);
            SIGNALS.as_ref()
        }
        
        fn constructed(&self) {
            let api_key_opt = self.application.borrow().api_key();
            if let Some(api_key) = api_key_opt {
                self.api_key_field.set_text(api_key.as_str());
            }
        }
    }

    impl WidgetImpl for SettingsPanel {
        fn map(&self) {
            self.parent_map();
            let main_loop = glib::MainContext::default();
            main_loop.spawn_local(clone!(@weak self as this => async move {
                this.set_is_loading(true);
                if !this.refresh_user_info().await {
                    this.refresh_stats().await;
                }
                this.set_is_loading(false);
                this.content_wrapper.get().set_visible(true);
            }));
        }
    }

    impl BoxImpl for SettingsPanel {}
}

glib::wrapper! {
    pub struct SettingsPanel(ObjectSubclass<imp::SettingsPanel>)
    @extends gtk::Widget, gtk::Box, @implements gtk::Accessible,
    gtk::Buildable, gtk::ConstraintTarget, gtk::Orientable;
}

impl SettingsPanel {
    pub fn new(app: &application::Application) -> Self {
        glib::Object::builder().property("application", app).build()
    }
}
