use gtk4 as gtk;
use gtk::{glib, gio, prelude::*, subclass::prelude::*};

mod imp {
    use crate::application;
    use super::*;
    use once_cell::sync::Lazy;
    use glib::{clone, subclass::Signal};
    
    #[derive(Debug, Default, gtk::CompositeTemplate)]
    #[template(file = "data/totp_window.blp")]
    pub struct TotpWindow {
        #[template_child]
        pub overlay: gtk::TemplateChild<gtk::Overlay>,
        #[template_child]
        pub form_wrapper: gtk::TemplateChild<gtk::Grid>,
        #[template_child]
        pub accept_button: gtk::TemplateChild<gtk::Button>,
        #[template_child]
        pub cancel_button: gtk::TemplateChild<gtk::Button>,
        #[template_child]
        pub code_entry: gtk::TemplateChild<gtk::Entry>,
        #[template_child]
        pub error_label: gtk::TemplateChild<gtk::Label>,
        #[template_child]
        pub loading_overlay: gtk::TemplateChild<gtk::Box>,
    }

    #[gtk::template_callbacks]
    impl TotpWindow {
        #[template_callback]
        fn totp_changed(&self, _: &gtk::Entry) {
            let totp = &self.code_entry.get().text();
            self.accept_button.get().set_sensitive(!totp.is_empty());
        }

        #[template_callback]
        fn canceled(&self, _: &gtk::Button) {
            self.obj().close();
        }

        #[template_callback]
        fn accepted(&self, _: &gtk::Button) {
            let main_loop = glib::MainContext::default();
            main_loop.spawn_local(clone!(@weak self as this => async move {
                let obj = this.obj();
                this.set_is_loading(true);
                this.loading_overlay.get().set_visible(true);
                let app: application::Application =
                obj.application().unwrap().downcast().unwrap();
                let mut ctx = app.context().borrow_mut();
                let totp = &this.code_entry.get().text();
                let mfa_res = ctx.mfa(totp).await;
                drop(ctx);
                match mfa_res {
                    Ok(()) => {
                        obj.emit_by_name::<()>("totp-success", &[]);
                        obj.close();
                    },
                    Err(e) if e.is_json() => this.set_error(e.to_string().as_str()),
                    Err(e) => {
                        obj.emit_by_name::<()>("totp-failed", &[&e.to_string()]);
                        obj.close();
                    },
                }
            }));
        }

        fn set_error(&self, error: &str) {
            self.set_is_loading(false);
            let error_label = &self.error_label.get();
            error_label.set_text(error);
            error_label.set_visible(true);
        }

        fn set_is_loading(&self, loading: bool) {
            if loading {
                self.error_label.get().set_visible(false);
            }
            self.loading_overlay.get().set_visible(loading);
            self.form_wrapper.get().set_sensitive(!loading);
        }
    }

    #[glib::object_subclass]
    impl ObjectSubclass for TotpWindow {
        const NAME: &'static str = "TotpWindow";
        type Type = super::TotpWindow;
        type ParentType = gtk::Window;

        fn class_init(class: &mut Self::Class) {
            class.bind_template();
            class.bind_template_callbacks();
        }

        fn instance_init(obj: &glib::subclass::InitializingObject<Self>) {
            obj.init_template();
        }
    }

    impl ObjectImpl for TotpWindow {
        fn signals() -> &'static [Signal] {
            static SIGNALS: Lazy<Vec<Signal>> = Lazy::new(|| 
                vec![Signal::builder("totp-success").build(),
                    Signal::builder("totp-failed")
                    .param_types([str::static_type()])
                    .build()]);
            SIGNALS.as_ref()
        }

        fn constructed(&self) {
            self.parent_constructed();
            self.overlay.add_overlay(&self.loading_overlay.get());
        }
    }

    impl WidgetImpl for TotpWindow {}
    impl WindowImpl for TotpWindow {}
}

glib::wrapper! {
    pub struct TotpWindow(ObjectSubclass<imp::TotpWindow>)
        @extends gtk::Widget, gtk::Window,
        @implements gio::ActionMap, gio::ActionGroup;
}

impl TotpWindow {
    pub fn new<P: IsA<gtk::Application>>(app: &P) -> Self {
        glib::Object::builder().property("application", app).build()
    }
}
